/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 * vim:expandtab:shiftwidth=8:tabstop=8:
 *
 * Copyright (C) 2001, 2002 Cluster File Systems, Inc.
 * Copyright (C) 2001 Tacit Networks, Inc. <phil@off.net>
 *
 *   This file is part of InterMezzo, http://www.inter-mezzo.org.
 *
 *   InterMezzo is free software; you can redistribute it and/or
 *   modify it under the terms of version 2 of the GNU General Public
 *   License as published by the Free Software Foundation.
 *
 *   InterMezzo is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with InterMezzo; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <glib.h>
#include <string.h>
#include <errno.h>
#include <netdb.h>
#include <sys/poll.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <fcntl.h>

#include <intersync.h>
#include <izolib.h>

struct is_peer *is_peer_set_ip(struct is_peer *peer, __u64 ip, __u64 port)
{
        struct in_addr ia;
        
        if (peer->srv_name != NULL)
                g_free(peer->srv_name);
        ia.s_addr = (unsigned int)ip;
        peer->srv_name = g_strdup(inet_ntoa(ia));
        peer->srv_ip = ip;
        peer->srv_port = port;

        peer->srv_addr.sin_family = AF_INET;
        peer->srv_addr.sin_port = htons(peer->srv_port);
        peer->srv_addr.sin_addr.s_addr = (unsigned int)ip;
        return peer;
}

struct is_peer *is_peer_new(__u64 ip, __u64 port,  __u64 my_port, int peer_type)
{
        struct is_peer *peer;
        
        peer = g_new0(struct is_peer, 1);
        CDEBUG(D_INFO, "setting up new peer %llx port %llu\n", ip, port);
        is_peer_set_ip(peer, ip, port);

        peer->srv_peer_type = peer_type;
        peer->srv_local_port = my_port;
        peer->srv_kml_mutex = g_mutex_new();
        peer->srv_conn_mutex = g_mutex_new();
        peer->srv_go_fetch_pending = 0;

        peer->srv_cxn_mutex = g_mutex_new();
        peer->srv_cxn_cond = g_cond_new();
        peer->srv_cxn_thread = NULL;
        peer->srv_connected = 0;

        INIT_LIST_HEAD(&peer->srv_conn_list);
        INIT_LIST_HEAD(&peer->srv_fset_list);
        INIT_LIST_HEAD(&peer->srv_replicator_list);
        return peer;
}

struct is_peer *is_peer_uuid_lookup(struct is_cell *cell, uuid_t uuid)
{
        struct list_head *tmp;

        list_for_each(tmp, &cell->cell_peer_list) {
                struct is_peer *server = list_entry(tmp, struct is_peer,
                                                    srv_servers_item);
                if (memcmp(server->srv_uuid, uuid,
                           sizeof(server->srv_uuid)) == 0)
                        return server;
        }
        return NULL;
}

struct is_peer *is_peer_ip_lookup(struct is_cell *cell, __u64 ip, __u64 port)
{
        struct list_head *tmp;

        list_for_each(tmp, &cell->cell_peer_list) {
                struct is_peer *server = list_entry(tmp, struct is_peer,
                                                      srv_servers_item);
                if (server->srv_ip == ip && server->srv_port == port)
                        return server;
        }

        return NULL;
}

int is_lowlevel_connect(struct sockaddr *addr, int nonblock)
{
        struct pollfd pollfd[1];
        int rc;

        pollfd[0].fd = is_socket_new(nonblock);

        if (connect(pollfd[0].fd, addr, sizeof(*addr)) < 0) {
                if (errno != EINPROGRESS) {
                        close(pollfd[0].fd);
                        return -1;
                }
        }

        if (nonblock) {
                int err, len;
                pollfd[0].events = POLLIN | POLLOUT | POLLERR | POLLHUP;
                rc = poll(pollfd, 1, 100);
                if (rc < 0) {
                        CERROR("poll() failed: %s\n", g_strerror(errno));
                        close(pollfd[0].fd);
                        return -1;
                } else if (rc == 0) {
                        close(pollfd[0].fd);
                        return -ETIMEDOUT;
                }
                getsockopt(pollfd[0].fd, SOL_SOCKET, SO_ERROR, &err, &len);
                if (err != 0) {
                        errno = err;
                        return -1;
                }
        }

        return pollfd[0].fd;
}

struct is_connection * is_get_connection(struct is_peer *server, int *err)
{
        int  junk, *error;
        struct is_connection *conn = NULL;

        if (err == NULL)
                error = &junk;
        else
                error = err;

        g_mutex_lock(server->srv_conn_mutex);
        if (!list_empty(&server->srv_conn_list)) {
                conn = list_entry(server->srv_conn_list.next,
                                  struct is_connection, conn_srv_item);
                list_del(&conn->conn_srv_item);
                goto out;
        }

        conn = g_new0(struct is_connection, 1);
        if (!conn) {
                CERROR("cannot get new is_conn");
                *error = -ENOMEM;
                goto out;
        }
        INIT_LIST_HEAD(&conn->conn_srv_item);
        conn->conn_client = httpclient_alloc();
 out:
        g_mutex_unlock(server->srv_conn_mutex);
        return conn;
}

void is_put_connection(struct is_peer *server, struct is_connection *conn)
{
        httpclient_reset(conn->conn_client);
        INIT_LIST_HEAD(&conn->conn_srv_item);

        g_mutex_lock(server->srv_conn_mutex);
        list_add(&conn->conn_srv_item, server->srv_conn_list.prev);
        g_mutex_unlock(server->srv_conn_mutex);
}

void is_connection_destroy(struct is_connection *conn)
{
        httpclient_free(conn->conn_client);
        g_free(conn);
}

void is_connection_destroy_all(struct is_peer * server)
{
        struct is_connection * conn;

        while (!list_empty(&server->srv_conn_list)) {
                conn = list_entry(server->srv_conn_list.next,
                                  struct is_connection, conn_srv_item);
                list_del(&conn->conn_srv_item);
                is_connection_destroy(conn);
        }
}
