/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 * vim:expandtab:shiftwidth=8:tabstop=8:
 *
 * Copyright (C) 2001, 2002 Cluster File Systems, Inc.
 *   Author: Phil Schwan <phil@clusterfs.com>
 *   Author: Peter J. Braam <braam@clusterfs.com>
 *
 *   This file is part of InterMezzo, http://www.inter-mezzo.org.
 *
 *   InterMezzo is free software; you can redistribute it and/or
 *   modify it under the terms of version 2 of the GNU General Public
 *   License as published by the Free Software Foundation.
 *
 *   InterMezzo is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with InterMezzo; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <grp.h>
#include <pwd.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>
#include <glib.h>

#include <stdio.h>
#include <signal.h>

#define _GNU_SOURCE
#include <getopt.h>
#undef _GNU_SOURCE

#include <intersync.h>
#include <izolib.h>

#include <signal.h>
#include <string.h>

# ifdef INTERSYNC_DEBUG
/* global variables */
int intersync_debug = 0;
int intersync_print_entry =0;
long intersync_kmemory = 0;
long intersync_vmemory = 0;
#endif

struct intersync *is;

static void create_docroot(char *docroot, char *cache)
{
        struct stat statbuf;
        char *tgt;

        if (stat(docroot, &statbuf) != 0) {
                if (mkdir(docroot, 0755) != 0)
                        g_error("Unable to mkdir \"%s\": %s\n",
                                docroot, g_strerror(errno));
        } else if (!S_ISDIR(statbuf.st_mode))
                g_error("Docroot \"%s\" exists and is not a directory.\n",
                        docroot);

        tgt = g_strdup_printf("%s/cache", docroot);
        /* remove old link, just in case mount point has changed */
        if (unlink(tgt) != 0 && errno != ENOENT) {
                g_error("Cannot remove previous symlink %s: %s",
                        tgt, g_strerror(errno));
        }                
        if (symlink(cache, tgt) != 0 && errno != EEXIST) {
                g_error("Cannot symlink %s->%s: %s",
                        tgt, cache, g_strerror(errno));
        }
        g_free(tgt);
}


static void kill_pidfile(char *fname, int sig)
{
        FILE *fp = fopen(fname, "r");
        if (fp) {
                pid_t pid;
                if (fscanf(fp, "%u", &pid) == 1 && kill(pid, sig))
                        CERROR("Cannot kill %d with signal %d: %s\n",
                                  pid, sig, strerror(errno));
                fclose(fp);
        } else if (errno != ENOENT)
                CERROR("Cannot read pid file `%s': %s\n",
                          fname, strerror(errno));
}


static char *pidfile = NULL;
static void signal_server(int sig)
{
        if (pidfile)
                kill_pidfile(pidfile, sig);

        if (sig == SIGHUP)
                CERROR("Reloading apache");
        else {
                pidfile = NULL;
                g_error("Shutting down on signal %d\n", sig);
        }
}


static void start_server(char *docroot, char * ip, int port)
{
        struct passwd *pw; 
        struct group *gr;
        struct stat stbuf;
        struct sigaction sigact;
        char *cmd, *httpd, *configfile;
        int rc;
        struct sockaddr_in sa;
        __u64 localip;

        gr = getgrnam("intermezzo");
        if (!gr) {
                system("/usr/sbin/groupadd intermezzo"); 
                gr = getgrnam("intermezzo"); 
                if (!gr) 
                        g_error("There is no 'intermezzo' group, and "
                                "'groupadd' failed.\n");
        }

        pw = getpwnam("intermezzo"); 
        if (!pw) { 
                system("/usr/sbin/useradd -g intermezzo -s /bin/true -d "
                       "/dev/null intermezzo");
                pw = getpwnam("intermezzo");
                if (!pw) 
                        g_error("There is no 'intermezzo' user, and 'useradd' "
                                "failed.\n");
        }

        pidfile = g_strdup_printf("%s/httpd.pid", docroot);
        configfile = g_strdup_printf("%s/intersync-httpd.conf", docroot);
        config_apache_conf(configfile, ip, port, docroot, IS_CGIDIR);

        sigact.sa_handler = signal_server;
        sigfillset(&sigact.sa_mask);
        sigact.sa_flags = SA_RESTART;
        sigaction(SIGHUP, &sigact, NULL);
        sigaction(SIGINT, &sigact, NULL);
        sigaction(SIGTERM, &sigact, NULL);

        httpd = "/usr/sbin/httpd";
        if (stat(httpd, &stbuf) && errno == ENOENT)
                httpd = "/usr/sbin/apache";
        cmd = g_strdup_printf("%s -f %s", httpd, configfile);

        rc = system(cmd);
        g_free(cmd);
        if (rc)
                g_error("Cannot start apache with config file %s: error code "
                        "0x%x\n", configfile, rc);
        g_free(configfile);

        /* Wait until Apache is capable of accepting connections before
         * returning. */

        sa.sin_family = AF_INET;
        sa.sin_port = htons(port);

        if (strcmp(ip, DEFAULT_LOCAL_IP) == 0 || izo_resolve(ip, &localip) < 0) {
                sa.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
        } else {
                sa.sin_addr.s_addr = (unsigned int)localip;
        }
        while (1) {
                rc = is_lowlevel_connect((struct sockaddr *)&sa, 0);
                if (rc < 0 && errno == ECONNREFUSED)
                        continue;
                if (rc < 0)
                        g_error("Cannot connect to Apache: %s\n",
                                g_strerror(errno));
                break;
        }
        close(rc);
}

static void usage(char *argv0, int status)
{
        fprintf(status ? stderr : stdout,
"Usage: %s [OPTION...] <cache directory>\n\
\n\
-d, --docroot=DIRECTORY   Set the docroot that contains the cache\n\
-i, --local-ip=IP         Local listen ip for http (default *)\n\
-p, --local-port=PORT     Local listen port (default %d or %d for clients)\n\
\n\
    --server=HOSTNAME     The server to contact for the initial profile\n\
    --server-port=PORT    Server connect port (default %d)\n\
\n\
    --clientid=CLIENTID   Configure system using CLIENTID\n\
    --fset=FSETNAME       Use fsetname, can include branch\n\
    --help                display this help and exit\n\
    --debug=FLAGS         print debug information\n\
    --hoard=FILE          hoard the files listed in FILE\n\
    --reint               quit after reintegration and hoard is done\n\
    --daemon              run intersync as daemon process\n\
    --version             output version information and exit\n\
\n\
  Report bugs to <intermezzo-devel@lists.sourceforge.net>\n",
                argv0, DEFAULT_SERVER_PORT, DEFAULT_CLIENT_PORT,
                DEFAULT_SERVER_PORT);

        exit(status);
}

static int get_args(int argc, char *argv[], struct is_config *cfg)
{
        int c;
        char *short_opts = "d:i:p:";
        static struct option long_opts[] = {
#define OPT_SERVER -2
                {"server", required_argument, NULL, OPT_SERVER},
#define OPT_SERVER_PORT -3
                {"server-port", required_argument, NULL, OPT_SERVER_PORT},
#define OPT_LOCAL_IP 'i'
                {"local-ip", required_argument, NULL, OPT_LOCAL_IP},
#define OPT_LOCAL_PORT 'p'
                {"local-port", required_argument, NULL, OPT_LOCAL_PORT},
#define OPT_DOCROOT 'd'
                {"docroot", required_argument, NULL, OPT_DOCROOT},
#define OPT_HELP -4
                {"help", no_argument, NULL, OPT_HELP},
#define OPT_VERSION -5
                {"version", no_argument, NULL, OPT_VERSION},
#define OPT_DEBUG -6
                {"debug", required_argument, NULL, OPT_DEBUG},
#define OPT_HOARD -7
                {"hoard", required_argument, NULL, OPT_HOARD},
#define OPT_REINT -8
                {"reint", no_argument, NULL, OPT_REINT},
#define OPT_CLIENTID -9
                {"clientid", required_argument, NULL, OPT_CLIENTID},
#define OPT_FSET -10
                {"fsetname", required_argument, NULL, OPT_FSET},
#define OPT_DAEMON -11
                {"daemon", no_argument, NULL, OPT_DAEMON},
                {0}
        };

        memset(cfg, 0, sizeof(*cfg));
        cfg->cfg_docroot = NULL;
        cfg->cfg_server_port = -1;
        cfg->cfg_local_ip = NULL;
        cfg->cfg_local_port = -1;
        cfg->cfg_hoard_conf = NULL;
        cfg->cfg_reint_only = 0;
        cfg->cfg_daemon = 0;
        cfg->cfg_clientid = NULL;
        cfg->cfg_fsetname = NULL;

        do {
                c = getopt_long(argc, argv, short_opts, long_opts, NULL);

                switch (c) {
                case OPT_HELP:
                        usage(argv[0], 0);
                        break;
                case OPT_VERSION:
                        fputs(PACKAGE " " VERSION "\n", stdout);
                        exit(0);
                        break;
                case OPT_SERVER:
                        cfg->cfg_server = g_strdup(optarg);
                        break;
                case OPT_SERVER_PORT:
                        cfg->cfg_server_port = atol(optarg);
                        break;
                case OPT_DOCROOT:
                        cfg->cfg_docroot = g_strdup(optarg);
                        break;
                case OPT_LOCAL_IP:
                        cfg->cfg_local_ip = g_strdup(optarg);
                        break;
                case OPT_LOCAL_PORT:
                        cfg->cfg_local_port = atol(optarg);
                        break;
                case OPT_DEBUG:
                        intersync_debug = atol(optarg);
                        break;
                case OPT_HOARD:
                        cfg->cfg_hoard_conf = g_strdup(optarg);
                        break;
                case OPT_REINT:
                        cfg->cfg_reint_only = 1;
                        break;
                case OPT_DAEMON:
                        cfg->cfg_daemon = 1;
                        break;
                case OPT_CLIENTID:
                        cfg->cfg_clientid = g_strdup(optarg);
                        break;
                case OPT_FSET:
                        cfg->cfg_fsetname = g_strdup(optarg);
                        break;
                case '?':
                        usage(argv[0], 1);
                        break;
                }
        } while (c != -1);

        if (optind != argc - 1) {
                /* We need exactly one non-option argument */
                usage(argv[0], 1);
        }
        cfg->cfg_cache = g_strdup(argv[optind]);
        return 0;
}

static struct is_config * 
get_profile(char *clientid, char * server, int server_port) 
{
        struct is_config *prof;
        char * url;
        struct httpclient *client;
        int len;
        char * buf;
        int rc;

        if (clientid) {
                url = g_strdup_printf("http://%s:%d/cgi-bin/cgi_profile?clientid=%s",
                                      server, server_port, clientid);
        } else {
                url = g_strdup_printf("http://%s:%d/cgi-bin/cgi_profile",
                                      server, server_port);
        }
        CDEBUG(D_INFO, "GET %s\n", url);

        /* FIXME: how about a little error checking in here, eh? */
        client = httpclient_alloc();
        CDEBUG(D_INFO, "httpclient_fetch(%p, %s, 0, 0)\n", client, url);
        rc = httpclient_fetch(client, url, 0, 0);
        if (rc) {
                httpclient_free(client);
                return NULL;
        }
        len = client->len;
        buf = g_malloc(len+1);
        memcpy (buf, client->body, len);
        buf[len] = '\0';
        httpclient_free(client);
        prof = is_parse_response(buf);

        g_free(url);
        g_free(buf);
        return prof;
}


int new_uuid(struct is_config *cfg) 
{
        uuid_t uuid;
        char uuidstr[UUID_UNPARSED_SIZE];
        uuid_generate(uuid);
        uuid_unparse(uuid, uuidstr);
        cfg->cfg_uuid = g_strdup(uuidstr);
        return 0;
}

/*
 * Copy config info to dst if not already set.
 */
int cfg_set(struct is_config *dst, struct is_config *src)
{
        if (!dst->cfg_uuid)
                dst->cfg_uuid = src->cfg_uuid;
        if (!dst->cfg_fsetname)
                dst->cfg_fsetname = src->cfg_fsetname;
        if (!dst->cfg_server)
                dst->cfg_server = src->cfg_server;
        if (dst->cfg_server_port == -1)
                dst->cfg_server_port = src->cfg_server_port;
        if (dst->cfg_local_port == -1)
                dst->cfg_local_port = src->cfg_local_port;
        if (!dst->cfg_local_ip) {
                dst->cfg_local_ip = src->cfg_local_ip;
        }
        return 0;
}

int cfg_defaults(struct is_config *cfg) 
{
        if (cfg->cfg_server != NULL && cfg->cfg_server[0] == '\0')
                cfg->cfg_server = NULL;
        if (cfg->cfg_server_port == -1)
                cfg->cfg_server_port = DEFAULT_SERVER_PORT;
        if (cfg->cfg_local_port == -1) {
                if (cfg->cfg_server != NULL) 
                        cfg->cfg_local_port = DEFAULT_CLIENT_PORT;
                else
                        cfg->cfg_local_port = DEFAULT_SERVER_PORT;
        }
        if (!cfg->cfg_local_ip) {
                cfg->cfg_local_ip = g_strdup(DEFAULT_LOCAL_IP);
        }
        if (!cfg->cfg_fsetname && !cfg->cfg_server)
                cfg->cfg_fsetname = g_strdup("rootfset");
        return 0;
}

/*
 * Normalize config parameters.
 *
 * - Remove trailing '/' from cache 
 */
int cfg_sanitize(struct is_config *cfg) 
{
        int len;

        len = strlen(cfg->cfg_cache);
        if (len > 1 && cfg->cfg_cache[len-1] == '/')
                cfg->cfg_cache[len-1] = '\0';
        return 0;
}

/*
 * Configuration details can  from 3 sources:
 *   1 command line
 *   2 cache config file
 *   3 remote profile 
 */
int is_init_config(int argc, char **argv, struct is_config *cfg) 
{
        int rc;
        struct is_config *tmp_cfg;

        get_args(argc, argv, cfg);

        if (!cfg->cfg_cache)
                usage(argv[0], 1);

        tmp_cfg = read_local_config(cfg->cfg_cache);
        if (tmp_cfg) 
                cfg_set(cfg, tmp_cfg);

        cfg_defaults(cfg);
        
        if (!cfg->cfg_fsetname && cfg->cfg_server) {
                tmp_cfg = get_profile(cfg->cfg_clientid, cfg->cfg_server, 
                                      cfg->cfg_server_port);
                if (tmp_cfg == NULL || tmp_cfg->cfg_fsetname == NULL)
                        return -1;
                cfg_set(cfg, tmp_cfg);
        }
        
        if (!cfg->cfg_uuid) 
                new_uuid(cfg);

        cfg_sanitize(cfg);
        
        rc = write_local_config(cfg);
        if (rc)
                is_die("unable to save local profile\n");
        return 0;
}

int logger_pipe_open(char *tag)
{
   int pid, p[2];
   pipe(p);

   if((pid = fork()) == -1)
   {
      fprintf(stderr, "fork failed\n");
      return -1;
   }
   else if(pid == 0)
   {
      close(p[1]);
      dup2(p[0], STDIN_FILENO);
      close(p[0]);
      execlp("logger", "logger", "-t", tag, NULL);
      fprintf(stderr, "exec failed\n");
      return -1;
   }
   close(p[0]);

   return p[1];
}

void daemonize(int channel)
{
   pid_t pid;
   int logger_fd;
   char tag[1024];

  /* Become a demon. */
   if( (pid = fork() ) < 0)
   {
        perror("Cannot fork");
        exit(-1);
   }

   if( pid > 0) /* Parent exits. */
   {
        FILE *pidfile;
        char pidfilename[26];
        sprintf(pidfilename, "/var/run/intersync-%d.pid", channel);
        pidfile = fopen(pidfilename, "w");
        fprintf(pidfile, "%d\n", pid);
        fclose(pidfile);
        exit(0);
   }

   
   /* Child continues. */
   setsid();
   chroot("/");

   snprintf(tag, sizeof(tag), "intersync-%d", channel);
   logger_fd = logger_pipe_open(tag);
   //close(1);
   dup2(logger_fd, 1);
   //close(2);
   dup2(logger_fd, 2);
   close(logger_fd);
   close(0);
}

int main(int argc, char **argv)
{
        struct is_peer *server = NULL;
        struct is_fileset *fset;
        struct is_config cfg;
        struct is_cache *cache;
        struct is_cell *cell;
        GMainLoop *main_loop;
        struct sigaction sigact;
        int rc;

        sigact.sa_handler = SIG_IGN;
        sigemptyset(&sigact.sa_mask);
        sigact.sa_flags = SA_RESTART;
        sigaction(SIGPIPE, &sigact, NULL);

        /* FIXME: turn on line buffering */
        setlinebuf(stdout);
        setlinebuf(stderr);

        rc = is_init_config(argc, argv, &cfg);
        if (rc)
                is_die("unable to determine configuration\n");

        is = is_intersync_new(cfg.cfg_uuid);

        /* XXX caches should be detected from /etc/fstab or from profile */
        cache = is_cache_initialize(is, cfg.cfg_cache, cfg.cfg_docroot);
        if (cache == NULL) 
                is_die("Unable to initalize cache %s\n", cfg.cfg_cache);

        if (cfg.cfg_local_port > 0) {
                create_docroot(cache->cache_docroot, cache->cache_mtpt);
                start_server(cache->cache_docroot, cfg.cfg_local_ip, cfg.cfg_local_port);
        }

	if (cfg.cfg_daemon == 1) {
		daemonize(cache->cache_channel);
	}

        rc = is_init_kernel(cache);
        if (rc)
                is_die("unable to initialize primary cache: %s\n", cache->cache_mtpt);
        rc = is_init_session(is);
        if (rc)
                is_die("unable to initialize session threads: %d\n", rc);

        /* XXX note that we're associating cell names with hostnames */
        cell = cell_new(is, cfg.cfg_server);

        /* get 'server' from 'servername' */
        if (cfg.cfg_server) {
                server = cell_add_server(cell, cfg.cfg_server,
                                         cfg.cfg_server_port,
                                         cfg.cfg_local_port);
                if (server == NULL)
                        g_error("Cannot connect to server.\n");
        }
        fset = is_fileset_initialize(cache, server, cfg.cfg_fsetname);
        if (!fset)
                is_die("failed to get root file set");

#if 0
        if(cfg.cfg_hoard_conf != NULL) 
               is_fileset_hoard(fset->fset_mtpt, cfg.cfg_hoard_conf);

        if (cfg.cfg_reint_only == 1) {
                signal_server(15);
                exit(1);        /* is there a g_exit()? */
        }
#endif

        main_loop = g_main_new(0);
        g_main_run(main_loop);

        return 0;
}


