/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 * vim:expandtab:shiftwidth=8:tabstop=8:
 *
 * Copyright (C) 2001, 2002 Cluster File Systems, Inc.
 * Copyright (C) 2001 Tacit Networks, Inc. <phil@off.net>
 *
 *   This file is part of InterMezzo, http://www.inter-mezzo.org.
 *
 *   InterMezzo is free software; you can redistribute it and/or
 *   modify it under the terms of version 2 of the GNU General Public
 *   License as published by the Free Software Foundation.
 *
 *   InterMezzo is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with InterMezzo; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <glib.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>

#include <asm/types.h>
#include <linux/intermezzo_fs.h>

#include "intersync.h"
#include <izolib.h>

/* Local upc handling functions */
static int izo_handle_upc_connect(struct intersync *is,struct intersync_upcall *upc);

static gboolean upc_prepare(GSource *source, gint *timeout)
{
        *timeout = -1; /* block forever */
        return FALSE; /* poll to find out if we're ready. */
}

static gboolean upc_check(GSource *source)
{
        GSList *l;

        for (l = source->poll_fds; l != NULL; l = l->next) {
                GPollFD *pfd = (GPollFD *)l->data;

                if (pfd->revents != 0)
                        return 1;
        }
        return 0;
}

static void upc_reply(GPollFD *pfd, struct intersync_upcall *upc, int rc)
{
        struct izo_upcall_resp reply;
        int tmp;

        reply.opcode = upc->u_hdr->u_opc;
        reply.unique = upc->u_hdr->u_uniq;
        reply.result = rc;

        CDEBUG(D_UPCALL, "Sending reply to Upcall (rc = %d)...\n", rc);
        tmp = 0;
        while (tmp < sizeof(reply)) {
                rc = write(pfd->fd, &reply, sizeof(reply));
                if (rc < 0)
                        break;
                tmp += rc;
        }
        if (tmp != sizeof(reply)) {
                CERROR("error writing upcall reply: %s\n", g_strerror(errno));
                CERROR("This should never, ever happen\n");
        }
}

static gboolean upc_dispatch(GSource *source, GSourceFunc callback,
                             gpointer data)
{
        struct intersync *is = data;
        GPollFD *pfd = &is->is_kernel_pfd;
        char *buf, *tmp, *ptr;
        struct intersync_upcall upc;
        int len, rc;

        buf = ptr = g_malloc(8192); 
        /* XXX fix for pipe. The kernel gives presently exactly one upcall, 
           not with pipes */
        len = read(pfd->fd, ptr, 8192);
        if (len < 0) {
                //CERROR("read from kernel failed: %s\n", g_strerror(errno));
                goto error;
        }
        CDEBUG(D_CHAT, "UPC: len %d\n", len);

        rc = upcall_unpack(&upc, &ptr, ptr + len);
        if (rc < 0) {
                CERROR("upcall_unpack failed: %s\n", g_strerror(-rc));
                goto error;
        }

        tmp = upcall_print(&upc, 1);
        CDEBUG(D_UPCALL, "%s", tmp);
        g_free(tmp);

        if (upc.u_hdr->u_version != IZO_UPC_VERSION) {
                g_error("Upcall version %0x, expected %0x!\n",
                        upc.u_hdr->u_version, IZO_UPC_VERSION);
                /* Not reached */
        }

        switch (upc.u_hdr->u_opc) {
        case IZO_UPC_OPEN: {
                /* Data on demand */
                struct is_fileset *fset;
                struct is_backfetch_data *bf_data;
                struct lento_vfs_context *info = 
                        (struct lento_vfs_context *)upc.u_rec;

                fset = is_fileset_lookup(is, MKSTR(upc.u_fset));
                if (!fset) {
                        CERROR("upcall dropped for unknown fileset \"%s\"\n",
                                  MKSTR(upc.u_fset));
                        goto error;
                }

                bf_data = g_new0(struct is_backfetch_data, 1);
                bf_data->info = info;
                bf_data->path = upc.u_path;

                rc = is_run_job(is, CONTEXT_JOB_BACKFETCH, NULL, 
                                fset->fset_server, bf_data);
                upc_reply(pfd, &upc, rc);
                break;
        }

        case IZO_UPC_GET_FILEID: {
                /* fetch by fileid */
                struct is_fileset *fset;
                struct is_fileid_data *fid_data;

                fset = is_fileset_lookup(is, MKSTR(upc.u_fset));
                if (!fset) {
                        CERROR("upcall dropped for unknown fileset \"%s\"\n",
                                  MKSTR(upc.u_fset));
                        goto error;
                }
                fid_data = g_new0(struct is_fileid_data, 1);
                fid_data->path = upc.u_path;
                fid_data->rec = upc.u_rec;
                fid_data->reclen = upc.u_hdr->u_reclen;
                rc = is_run_job(is, CONTEXT_JOB_GET_FILEID, NULL, 
                                fset->fset_server, fid_data);
                        
                upc_reply(pfd, &upc, rc);
                break;
        }

        case IZO_UPC_BACKFETCH: {
                /* Similar to data-on-demand, but done on the server.  Since the
                 * server doesn't know which client the update came from, the
                 * UUID is passed in as part of the upcall. */
                struct is_fileset *fset;
                struct is_peer *peer;
                struct is_replicator *rep;
                struct is_backfetch_data *bf_data;
                struct lento_vfs_context *info = 
                        (struct lento_vfs_context *)upc.u_rec;
                rc = -EINVAL;

                fset = is_fileset_lookup(is, MKSTR(upc.u_fset));
                if (!fset) {
                        CERROR("upcall dropped for unknown fileset \"%s\"\n",
                                  MKSTR(upc.u_fset));
                        goto bf_error;
                }

                peer = is_peer_uuid_lookup(is_get_cell(is, NULL), info->uuid);
                if (peer == NULL) {
                        char uuidstr[UUID_UNPARSED_SIZE];
                        uuid_unparse(upc.u_hdr->u_uuid, uuidstr);
                        CERROR("Received BACKFETCH from unknown peer "
                                  "(uuid %s): dropping.\n", uuidstr);
                        rc = -ENOTCONN;
                        goto bf_error;
                }

                rep = is_replicator_lookup(fset, peer);
                if (rep == NULL) {
                        char uuidstr[UUID_UNPARSED_SIZE];
                        uuid_unparse(upc.u_hdr->u_uuid, uuidstr);
                        CERROR("Received BACKFETCH for peer with no "
                                  "replicator (uuid %s): dropping.\n", uuidstr);
                        goto bf_error;
                }

                bf_data = g_new0(struct is_backfetch_data, 1);
                bf_data->path = upc.u_path;
                bf_data->info = info; 

                rc = is_run_job(is, CONTEXT_JOB_BACKFETCH, NULL, rep, bf_data);
        bf_error:
                upc_reply(pfd, &upc, rc);
                break;
        }
        case IZO_UPC_CONNECT: {
                /* Initiate a connection */
                rc = izo_handle_upc_connect(is, &upc);

                upc_reply(pfd, &upc, rc);
                break;
        }
        case IZO_UPC_GO_FETCH_KML: {
                /* On top of the obvious purpose of this request (this peer is
                 * being asked to fetch new KML from the sending peer), it is
                 * also the first request issued by a peer after it sends a
                 * uuid-with-reverse (CONNECT upcall) to the server.  In this
                 * way, the server learns which filesets this peer is
                 * replicating and should create replicator structures to handle
                 * them. */
                struct is_fileset *fset;
                struct is_peer *sending_peer;
                struct is_replicator *rep;
                __u64 *kmlsize;

                fset = is_fileset_lookup(is, MKSTR(upc.u_fset));
                if (!fset) {
                        CERROR("upcall dropped for unknown fileset \"%s\"\n",
                                  MKSTR(upc.u_fset));
                        goto error;
                }

                sending_peer = is_peer_uuid_lookup(is_get_cell(is, NULL),
                                                   upc.u_hdr->u_uuid);

                if (sending_peer == NULL) {
                        char uuidstr[UUID_UNPARSED_SIZE];
                        uuid_unparse(upc.u_hdr->u_uuid, uuidstr);
                        CERROR("Received GO_FETCH_KML from unknown peer (%s)"
                                  ": dropping.\n", uuidstr);
                        goto error;
                }

                rep = is_replicator_lookup(fset, sending_peer);

                if (rep == NULL) {
                        is_die("replicator not found, should exist!\n");
                }

                kmlsize = g_malloc(sizeof(__u64));
                *kmlsize = upc.u_hdr->u_offset;
                rc = is_run_job(is, CONTEXT_JOB_KML, NULL, rep, kmlsize);
                break;
        }
        case IZO_UPC_PERMIT: {
                /* Get the permit from the current permit holder (which will
                 * always appear to be the server on a client) */
                /* Don't return until all available kml has been reintegrated. */
                struct is_fileset *fset;
                struct is_replicator *rep;

                fset = is_fileset_lookup(is, upc.u_fset);
                if (!fset) {
                        CERROR("upcall dropped for unknown fileset \"%s\"\n",
                                  MKSTR(upc.u_fset));


                        goto error;
                }

                if (fset->fset_permit_holder == NULL) {
                        /* We already have it! */
                        fset->fset_permit_holder = NULL;
                        upc_reply(pfd, &upc, 0);
                        break;
                }

                rep = is_replicator_lookup(fset, fset->fset_permit_holder);
                if (rep == NULL) {
                        CERROR("no replicator for permit holder?  very unexpected, tell phil.\n");
                        upc_reply(pfd, &upc, -ENOTCONN);
                        break;
                }

                rc = is_run_job(is, CONTEXT_JOB_GET_PERMIT, NULL, rep, NULL);

                if (rc == -ENOTCONN) {
                        CERROR("Disconnected, steal permit.\n");
                        fset->fset_permit_holder = NULL;
                        upc_reply(pfd, &upc, rc);
                        break;
                }
                if (rc != 0) {
                        CERROR("We failed to get the permit.  Do something more dramatic here.\n");
                        upc_reply(pfd, &upc, rc);
                        break;  /* this dramatic enough? */
                }

                if (fset->fset_server == NULL) {
                        /* If we're the server, tell the permit holder what the
                         * new local KML size is so that it knows to skip KML
                         * that was generated as a result of records that were
                         * reintegrated while it held the permit. */
                        rc = is_run_job(is, CONTEXT_JOB_SEND_KMLSIZE, NULL, rep,
                                        NULL);
                }
                if (rc == 0) {
                        is_wait_for_kml(fset);
                        fset->fset_permit_holder = NULL;
                }
                upc_reply(pfd, &upc, rc);
                break;
        }
        case IZO_UPC_REVOKE_PERMIT: {
                /* We are either the permit holder or the server.  Get the
                 * permit if we don't have it, and then give it to whoever is
                 * requesting it.  The kernel already knows that we're losing
                 * the permit, so there's no need to tell it again. */
                struct is_fileset *fset;
                struct is_peer *peer;
                struct is_replicator *rep;

                fset = is_fileset_lookup(is, upc.u_fset);
                if (!fset) {
                        CERROR("upcall dropped for unknown fileset \"%s\"\n",
                                  MKSTR(upc.u_fset));
                        goto error;
                }

                if (fset->fset_permit_holder != NULL) {
                        /* We're not the permit holder. */
                         if (fset->fset_server != NULL)  {
                                 /* we're not the server */
                                 /* we must be waiting for the permit ourselves */
                                 /* who's we, anyway? */
                                 upc_reply(pfd, &upc, -EINVAL);
                                 break;
                         }
                        rep = is_replicator_lookup(fset,
                                                   fset->fset_permit_holder);
                        if (rep == NULL) {
                                CERROR("no replicator for permit holder?  very unexpected, tell phil.\n");
                                upc_reply(pfd, &upc, -ENOTCONN);
                                break;
                        }

                        rc = is_run_job(is, CONTEXT_JOB_GET_PERMIT, NULL, rep,
                                        NULL);

                        if (rc == -ENOTCONN) {
                                CERROR("Disconnected from old permit holder, steal permit.\n");
                                fset->fset_permit_holder = NULL;
                        } else  if (rc) {
                                /* FIXME */
                                CERROR("We failed to get the permit.  Do something more dramatic here.\n");
                                upc_reply(pfd, &upc, rc);
                                break;
                        } else {
                                fset->fset_permit_holder = NULL;
                        }
                } 

                /* Get the peer of the new permit holder. */
                peer = is_peer_uuid_lookup(is_get_cell(is, NULL),
                                           upc.u_hdr->u_uuid);
                if (peer == NULL) {
                        CERROR("no peer for permit requestor?  very unexpected, tell phil.\n");
                        upc_reply(pfd, &upc, -ENOTCONN);
                        break;
                }

                CDEBUG(D_INFO, "Permit given to peer %p\n", peer);
                fset->fset_permit_holder = peer;
                upc_reply(pfd, &upc, 0);
                break;
        }
        case IZO_UPC_KML: {
                struct is_fileset *fset;
                struct list_head *lh;

                fset = is_fileset_lookup(is, upc.u_fset);
                if (!fset) {
                        CERROR("upcall dropped for unknown fileset \"%s\"\n",
                                  MKSTR(upc.u_fset));
                        goto error;
                }

                /* The server will get KML upcalls when it reintegrates KML from
                 * clients.  Clients should not. */
                if (fset->fset_permit_holder != NULL &&
                    fset->fset_server != NULL)
                        g_error("we're not the permit holder, but we got an IZO_UPC_KML?  That's strange.\n");

                /* Send a 'go_fetch_kml' request to any replicators. */
                list_for_each(lh, &fset->fset_replicator_list) {
                        __u64 *size = g_malloc(sizeof(__u64));
                        struct is_replicator *rep =
                                list_entry(lh, struct is_replicator,
                                           rp_fset_item);
                        {
                                char str[UUID_UNPARSED_SIZE];
                                uuid_unparse(rep->rp_peer->srv_uuid, str);
                                CDEBUG(D_KML, "holder: <%p> peer: <%p> %s:%d %s\n", 
                                       fset->fset_permit_holder,
                                       rep->rp_peer,
                                       rep->rp_peer->srv_name,
                                       rep->rp_peer->srv_port,
                                       str);
                        }

                        if (fset->fset_permit_holder == rep->rp_peer)
                                continue;

                        *size = upc.u_hdr->u_length;
                        is_go_fetch_notify(is, rep, size);
                }

                break;
        }
        case IZO_UPC_SET_KMLSIZE: {
                struct is_fileset *fset;

                fset = is_fileset_lookup(is, MKSTR(upc.u_fset));
                if (!fset) {
                        CERROR("upcall dropped for unknown fileset \"%s\"\n",
                                  MKSTR(upc.u_fset));
                        upc_reply(pfd, &upc, -EINVAL);
                        goto error;
                }

                if (fset->fset_server == NULL) {
                        CERROR("No server in UPC_SET_KMLSIZE?\n");
                        upc_reply(pfd, &upc, -EINVAL);
                        goto error;
                }

                fset->fset_server->rp_rcvd_rec.lr_remote_offset =
                        upc.u_hdr->u_length;
                upc_reply(pfd, &upc, 0);
                break;
        }

        case IZO_UPC_REPSTATUS: {
                struct is_fileset *fset;
                struct is_peer *peer;
                struct is_replicator *rep;
                struct izo_rcvd_rec *lr_peer;
                lr_peer = (struct izo_rcvd_rec *) upc.u_rec;
                if (!lr_peer) {
                        CERROR("No last_rcvd data\n");
                        upc_reply(pfd, &upc, -EINVAL);
                        goto error;

                }
                fset = is_fileset_lookup(is, MKSTR(upc.u_fset));
                if (!fset) {
                        CERROR("upcall dropped for unknown fileset \"%s\"\n",
                                  MKSTR(upc.u_fset));
                        upc_reply(pfd, &upc, -EINVAL);
                        goto error;
                }
                peer = is_peer_uuid_lookup(is_get_cell(is, NULL), lr_peer->lr_uuid);
                if (!peer) {
                        CERROR("upcall dropped for unknown peer\n");
                        upc_reply(pfd, &upc, -ENOTCONN);
                        goto error;
                }
                rep = is_replicator_lookup(fset, peer);
                if (!rep) {
                        CDEBUG(D_CHAT, "no replicator found.\n");
                        rep = is_replicator_new(fset, peer);
                }
                memcpy(&rep->rp_rcvd_rec, lr_peer, sizeof(rep->rp_rcvd_rec));
                upc_reply(pfd, &upc, 0);
                break;
        }

        case IZO_UPC_CLIENT_MAKE_BRANCH: {
                struct is_fileset *fset;
                char *tagname, *branchname;

                fset = is_fileset_lookup(is, MKSTR(upc.u_fset));
                if (!fset) {
                        CERROR("upcall dropped for unknown fileset \"%s\"\n",
                                  MKSTR(upc.u_fset));
                        upc_reply(pfd, &upc, -EINVAL);
                        goto error;
                }

                if (fset->fset_server == NULL) {
                        CERROR("No server in UPC_CLIENT_MAKE_BRANCH?\n");
                        upc_reply(pfd, &upc, -EINVAL);
                        goto error;
                }

                /* Path will contain "tagname/branchname" */
                tagname = upc.u_path;
                branchname = upc.u_rec;
                if ( *(branchname + upc.u_hdr->u_reclen -1) != '\0') {
                        g_error("upc packed wrongly -- see fs24/upcall.c\n");
                        goto error; 
                }

                if (branchname == NULL) {
                        CERROR("Bad upcall path in UPC_CLIENT_MAKE_BRANCH\n");
                        upc_reply(pfd, &upc, -EINVAL);
                        goto error;
                }

                g_error("implemente is_make_branch\n"); 

                /* send tag and branch */
                rc = is_run_job(is, CONTEXT_JOB_MAKE_BRANCH, NULL,
                                fset->fset_server, g_strdup(upc.u_path));

                upc_reply(pfd, &upc, 0);
                break;
        }

        case IZO_UPC_SERVER_MAKE_BRANCH: {
                struct is_fileset *fset;
                
                fset = is_fileset_lookup(is, MKSTR(upc.u_fset));
                if (fset) {
                        CERROR("fset exists \"%s\"\n",
                                  MKSTR(upc.u_fset));
                        upc_reply(pfd, &upc, -EINVAL);
                        goto error;
                }

                fset = is_fileset_make_branch(is, upc.u_fset);
                if (!fset) {
                        rc = -EINVAL;
                }
                upc_reply(pfd, &upc, rc);
                break;
        }

        case IZO_UPC_BRANCH_UNDO: {
                struct is_fileset *fset;

                fset = is_fileset_lookup(is, MKSTR(upc.u_fset));
                if (!fset) {
                        CERROR("upcall dropped for unknown fileset \"%s\"\n",
                                  MKSTR(upc.u_fset));
                        goto error;
                }

                rc = is_run_job(is, CONTEXT_JOB_BRANCH_UNDO, NULL,
                                fset->fset_server, g_strdup(upc.u_path));
                upc_reply(pfd, &upc, rc);
                break;
        }
        case IZO_UPC_BRANCH_REDO: {
                struct is_fileset *fset;

                fset = is_fileset_lookup(is, MKSTR(upc.u_fset));
                if (!fset) {
                        CERROR("upcall dropped for unknown fileset \"%s\"\n",
                                  MKSTR(upc.u_fset));
                        goto error;
                }

                rc = is_run_job(is, CONTEXT_JOB_BRANCH_REDO, NULL,
                                fset->fset_server, g_strdup(upc.u_path));
                upc_reply(pfd, &upc, rc);
                break;
        }
        default:
                CERROR("dropping unimplemented upcall\n");
                break;
        }

 error:
        g_free(buf);
        return TRUE;
}

GSourceFuncs upcall_source_funcs = {
        prepare:  upc_prepare,
        check:    upc_check,
        dispatch: upc_dispatch
};

void is_run_upcall(struct is_context *ctxt, gpointer data2)
{
        GSource *source;
        GMainLoop *main_loop;
        GPollFD *pfd = &ctxt->ctxt_intersync->is_kernel_pfd;
        GMainContext *gcontext;

        source = g_source_new(&upcall_source_funcs, sizeof(GSource));
        if (source == NULL) {
                g_error("Unable to allocate GSource\n");
                return;
        }

        gcontext = g_main_context_new();
        main_loop = g_main_loop_new(gcontext, TRUE);

        pfd->events = G_IO_IN | G_IO_HUP | G_IO_ERR;
        g_source_add_poll(source, pfd);
        g_source_set_priority(source, G_PRIORITY_DEFAULT);
        g_source_set_callback(source, NULL, ctxt->ctxt_intersync, NULL);

        /* FIXME: hack to work with older glib; is this correct? */
        g_source_attach(source, gcontext);

        g_main_run(main_loop);

        g_source_unref(source);
        g_main_loop_unref(main_loop);
}


/*
 * Handle connection requests from clients. Support:
 *  1. Bi-directional, fully-connected clients
 *  2. Readonly, half-connected clients (no httpd on client)
 *  3. Reconnect, which may change the connection mode 
 */
static int izo_handle_upc_connect(struct intersync *is,
                           struct intersync_upcall *upc) 
{
        struct is_peer *peer;
        __u64 ip, port;
        int peer_type;
        char uuidstr[UUID_UNPARSED_SIZE];
        int rc = 0;
                
        ip = upc->u_hdr->u_offset;
        port = upc->u_hdr->u_length;
        peer_type = upc->u_hdr->u_first_recno;

        uuid_unparse(upc->u_hdr->u_uuid, uuidstr);

        CDEBUG(D_UPCALL, "connect upcall for ip: %llx port: %llu " 
               "uuid: %s, client: %d\n", 
               ip, port, uuidstr, peer_type);

        if (peer_type != PEER_SERVER) 
                goto out;

        peer = is_peer_uuid_lookup(is_get_cell(is, NULL), 
                                   upc->u_hdr->u_uuid);
        if (peer == NULL) {
                peer = cell_new_client(is_get_cell(is, NULL),
                                       ip, port, 0,
                                     upc->u_hdr->u_uuid);
                if (peer == NULL) {
                        CERROR("Cannot create peer, out of memory\n");
                        rc = -ENOTCONN;
                        goto out;
                }
        } else {
                /* this peer was connected, and is now reconneting. 
                 * reset connection.
                 */
                CDEBUG(D_UPCALL, "Resetting peer: %s\n", peer->srv_name);
                is_connection_destroy_all(peer);
                is_release_all_permits(is, peer);
                is_peer_set_ip(peer, ip, port);
        }
        rc = is_confirm_connection(peer);

 out:
        return rc;
}
