/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 * vim:expandtab:shiftwidth=8:tabstop=8:
 *
 * Copyright (C) 2002  Andres Salomon <dilinger@voxel.net>
 *
 *   This file is part of InterMezzo, http://www.inter-mezzo.org.
 *
 *   InterMezzo is free software; you can redistribute it and/or
 *   modify it under the terms of version 2 of the GNU General Public
 *   License as published by the Free Software Foundation.
 *
 *   InterMezzo is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with InterMezzo; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <glib.h>
#include <izolib.h>
#include <errno.h>
#include <curl/curl.h>

/*
 * Synch regularly w/ curl.h.  Current version synched w/ curl 0.7.6.
 */
static char *curl_errors[] = {
        "Ok",
        "Unsupported protocol",
        "Failed initialization",
        "Malformed URL",
        "Malformed URL user",
        "Couldn't resolve proxy",
        "Couldn't resolve host",
        "Couldn't connect",
        "FTP server gave a weird reply",
        "FTP access denied",
        "FTP user or password is incorrect",
        "FTP weird password reply",
        "FTP weird user reply",
        "FTP weird pasv reply",
        "FTP weird 227 format",
        "FTP can't get host",
        "FTP can't reconnect",
        "FTP couldn't set binary mode",
        "Could only retrieve partial file",
        "FTP couldn't retrieve file",
        "FTP write error",
        "FTP quote error",
        "HTTP not found",
        "Write error",
        "Malformed user",
        "FTP couldn't upload file",
        "Read error",
        "Out of memory",
        "Operation timed out",
        "FTP couldn't set ascii mode",
        "FTP 'port' command failed",
        "FTP 'rest' command failed",
        "FTP 'size' command failed",
        "HTTP 'range' header failed",
        "HTTP POST error",
        "SSL connection error",
        "FTP can't resume download",
        "Can't read file",
        "LDAP cannot bind",
        "LDAP search failed",
        "Library not found",
        "Function not found",
        "Aborted by callback",
        "Bad function argument",
        "Bad calling order",
        "HTTP PORT failed",
        "Bad password entered",
        "Too many redirects",
        "Unknown telnet option",
        "Malformed telnet option",
        "Obsolete",
        "SSL peer certificate error",
        "Got nothing",
        "SSL engine not found",
        "SSL engine set failed",
        "Send error",
        "Receive error",
};
const int maxerrno = sizeof(curl_errors)/sizeof(curl_errors[0]);

/*
 * Convert curl errno's to strings.
 */
const char *httpclient_strerror(int rc)
{
        const char *s;

        if (rc < 0)
                s = "Invalid function argument";
        else if (rc < maxerrno)
                s = curl_errors[rc];
        else
                s = "Unknown/out-of-range error!";

        return s;
}

int httpclient_errno(int rc)
{
        int err = 0;

        if (rc == CURLE_OK)
                err = 0;
        else if (rc == CURLE_HTTP_NOT_FOUND)
                err = ENOENT;
        else if (rc == CURLE_COULDNT_CONNECT)
                err = ENOTCONN;
        else
                err = -1;      /* What should be the catchall? */

        return err;
}

/*
 * Callback for curl's data write.
 */
static size_t _body_save(void *ptr, size_t size, size_t nmemb, void *arg)
{
        struct httpclient *client = (struct httpclient *) arg;

        g_return_val_if_fail(client != NULL, 0);

        if (client->len) {
                /* append more data to body */
                size_t temp = client->len;

                client->len += size * nmemb;
                client->body = g_realloc(client->body, client->len);
                memcpy(client->body + temp, ptr, size * nmemb);
        }
        else {
                /* first time this function's been called */
                client->len = size * nmemb;
                client->body = g_malloc(client->len);
                memcpy(client->body, ptr, client->len);
        }

        return size * nmemb;
}

/*
 * Tell cURL to fetch 'uri', filling in the client struct w/ the body
 * and length of the query results.  httpclient_fetch_file(), below,
 * is faster than this function.  A non-zero return value indicates
 * failure.  Set 'start_byte' and 'end_byte' to 0 if you want to
 * fetch the entire body.
 */
int httpclient_fetch(struct httpclient *client, const char *uri,
                size_t start_byte, size_t end_byte)
{
        int rc;
        char *range = NULL;
        CURL *handle;
       
        g_return_val_if_fail(client != NULL && uri != NULL, -1);
        
        /* clean up residue from previous connections */
        if (client->len) {
                g_free(client->body);
                client->body = NULL;
                client->len = 0;
        }
        
        /* initialize curl object */
        if (!client->_handle)
                client->_handle = (void *) curl_easy_init();
        
        handle = (CURL *) client->_handle;
        rc = curl_easy_setopt(handle, CURLOPT_URL, uri);
        if (rc != CURLE_OK)
                goto done;
        rc = curl_easy_setopt(handle, CURLOPT_WRITEFUNCTION, _body_save);
        if (rc != CURLE_OK)
                goto done;
        rc = curl_easy_setopt(handle, CURLOPT_FILE, (void *) client);
        if (rc != CURLE_OK)
                goto done;
        rc = curl_easy_setopt(handle, CURLOPT_FAILONERROR, 1);
        if (rc != CURLE_OK)
                goto done;
        if (start_byte >= 0 && end_byte > 0) {
                range = g_strdup_printf("%d-%d", start_byte, end_byte);
                rc = curl_easy_setopt(handle, CURLOPT_RANGE, range);
                if (rc != CURLE_OK)
                        goto done;
        }
        else
                curl_easy_setopt(handle, CURLOPT_RANGE, "");
        
        /* do fetch */
        rc = curl_easy_perform(handle);

done:
        if (range)
                g_free(range);

        return rc;
}

/*
 * Same semantics as httpclient_fetch(), except instead of saving data into
 * the httpclient struct, save it directly to a file.  Except 'fp' to be
 * opened already, in some writable mode.  Due to the nature of this function,
 * you're probably better off writing to a temporary file, then
 * renaming to (overwriting) the target file, so that the contents of the
 * original are not truncated before ensuring that the fetch succeeded.
 */
int httpclient_fetch_file(struct httpclient *client, const char *uri,
                size_t start_byte, size_t end_byte, FILE *fp)
{
        int rc;
        char *range = NULL;
        CURL *handle;
        double x;
       
        g_return_val_if_fail(client != NULL && uri != NULL && fp != NULL, -1);
        
        /* clean up residue from previous connections */
        if (client->len) {
                g_free(client->body);
                client->body = NULL;
                client->len = 0;
        }
        
        /* initialize curl object */
        if (!client->_handle)
                client->_handle = (void *) curl_easy_init();
        
        handle = (CURL *) client->_handle;
        rc = curl_easy_setopt(handle, CURLOPT_URL, uri);
        if (rc != CURLE_OK)
                goto done;
        rc = curl_easy_setopt(handle, CURLOPT_WRITEFUNCTION, fwrite);
        if (rc != CURLE_OK)
                goto done;
        rc = curl_easy_setopt(handle, CURLOPT_FILE, (void *) fp);
        if (rc != CURLE_OK)
                goto done;
        rc = curl_easy_setopt(handle, CURLOPT_FAILONERROR, 1);
        if (rc != CURLE_OK)
                goto done;
        if (start_byte >= 0 && end_byte > 0) {
                range = g_strdup_printf("%d-%d", start_byte, end_byte);
                rc = curl_easy_setopt(handle, CURLOPT_RANGE, range);
                if (rc != CURLE_OK)
                        goto done;
        }
        else
                curl_easy_setopt(handle, CURLOPT_RANGE, "");
        
        /* do fetch */
        rc = curl_easy_perform(handle);
        if (rc != CURLE_OK)
                goto done;

        /* get length */
        curl_easy_getinfo(handle, CURLINFO_SIZE_DOWNLOAD, &x);
        client->len = (size_t) x;

done:
        if (range)
                g_free(range);

        return rc;
}

/*
 * Free io buffer, but don't close the connection.
 */
void httpclient_reset(struct httpclient *client)
{
        if (client && client->len) {
                g_free(client->body);
                client->body = NULL;
                client->len = 0;
        }
}

/*
 * Close connection and free all memory.
 */
void httpclient_free(struct httpclient *client)
{
        if (client) {
                httpclient_reset(client);
                if (client->_handle)
                        curl_easy_cleanup((CURL *) client->_handle);
                client->_handle = NULL;
                g_free(client);
        }
}
