/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 * vim:expandtab:shiftwidth=8:tabstop=8:
 *
 * Copyright (C) 2001 Cluster File Systems, Inc. <braam@clusterfs.com>
 *
 *   This file is part of InterMezzo, http://www.inter-mezzo.org.
 *
 *   InterMezzo is free software; you can redistribute it and/or
 *   modify it under the terms of version 2 of the GNU General Public
 *   License as published by the Free Software Foundation.
 *
 *   InterMezzo is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with InterMezzo; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * KML Unpacking 
 *
 */

#include <time.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>
#include <errno.h>
#include <sys/fsuid.h>
#include <grp.h>
#include <sys/time.h>
#include <utime.h>

#include <linux/intermezzo_lib.h>
#include <izolib.h>
#include <glib.h>


struct reint_data { 
        char *cwd;
        int ngroups;
        gid_t *groups;
        uid_t  fsuid;
        gid_t fsgid;
} fsreintdata;

#define out_errno(eno) do { rc = -eno; goto out; } while (0)
#define out_save_errno(code) do { int tmp = errno; code; out_errno(tmp); } while (0)

void init_fsreintdata (void)
{
        fsreintdata.fsuid = getuid();
        fsreintdata.fsgid = getgid();

        fsreintdata.groups = g_new(gid_t, NGROUPS_MAX);
        if ( !fsreintdata.groups ) { 
                g_error("out of memory");
        }
        
        fsreintdata.ngroups = getgroups(NGROUPS_MAX, fsreintdata.groups);
        if (fsreintdata.ngroups == -1) { 
                perror("cannot do getgroups\n");
                g_error("cannot do getgroups\n");
        }

        fsreintdata.cwd = getcwd(NULL, 0);
        if (!fsreintdata.cwd) { 
                perror("cannot get cwd");
                exit(1);
        }

}


static int reint_check_path(char *path)
{
        int len = strlen(path);
        if (len == 2 && !strcmp ("..", path))
                return -EPERM;
        if (!strncmp("../", path, 3))
                return -EPERM;
        if (len >= 3 && !strcmp("/..", path + len - 3))
                return -EPERM;
        if (strstr(path, "/../")) 
                return -EPERM;
        return 0;
}

static int reint_pre_secure(struct kml_rec *rec, char *mtpt)
{
        int rc = 0;
        gid_t groups[NGROUPS_MAX];
        int i;
        
        /* FIXME: These chdirs and chroots break things because we use
           makepath. */
        if ( geteuid() != 0 ) {
                /* chdir(mtpt); */
                return 0;
        }
#if 0
        if (rc = chdir(mtpt))
                out_save_errno(perror("pre_secure chdir"));
#endif
#if 0
        if (chroot(mtpt))
                out_save_errno(perror("pre_secure chroot"));
#endif

        memset(&groups, 0, sizeof(groups));
        for ( i=0 ;i<rec->prefix.hdr->ngroups;  i++  )
                groups[i] = rec->prefix.groups[i];

        if (setgroups(rec->prefix.hdr->ngroups, groups))
                out_save_errno(perror("reint_pre_secure"));

        if ( setfsuid(rec->prefix.hdr->fsuid) != rec->prefix.hdr->fsuid )
                out_save_errno(g_error("cannot fsuid (uid %d)",
                                       rec->prefix.hdr->fsuid));

        if ( setfsgid(rec->prefix.hdr->fsgid) != rec->prefix.hdr->fsgid )
                out_save_errno(g_error("cannot fsgid (gid %d)",
                                       rec->prefix.hdr->fsgid));

 out:
        return rc;
}

static void reint_post_secure (void) { 
        int err; 

        if ( geteuid() != 0 ) {
                return;
        }

        if ( (err = setgroups(fsreintdata.ngroups, fsreintdata.groups)) ) { 
                g_error("%s: setgroups: %s",
                        __FUNCTION__, g_strerror(errno)); 
        }

        if ( setfsuid(fsreintdata.fsuid) != fsreintdata.fsuid)  {
                g_error("cannot fsuid");
                exit(1);
        }

        if ( setfsgid(fsreintdata.fsgid) != fsreintdata.fsgid ) {
                g_error("cannot fsgid");
                exit(1);
        }

/*      if ( (err =  chdir(".") ) ) {  */
/*              perror("post_secure chdir ."); */
/*              exit(1); */
/*      } */

/*      if ( (err =  chroot(".") ) ) {  */
/*              perror("post_secure chroot"); */
/*              exit(1); */
/*      } */

        if ( (err =  chdir(fsreintdata.cwd) ) ) { 
                perror("post_secure chdir");
                exit(1);
        }

        return;
}       


static int setattr(char *path, int valid, int mode, int uid, int gid, long mtime, long ctime, loff_t size)
{
        int rc = 0;
        struct stat buf;
        struct timeval tv[2];

        if ( lstat(path, &buf) )
                out_save_errno(g_error("%s: setattr cannot stat: %s",
                                       __FUNCTION__, g_strerror(errno)));

        if (geteuid != 0)
                valid &= ~(ATTR_UID|ATTR_GID|ATTR_CTIME);

        if (valid & ATTR_MODE && !S_ISLNK(buf.st_mode) ) { 
                rc = chmod(path, mode);
                if (rc)
                        out_save_errno(g_error("%s: setattr chmod: %s",
                                               __FUNCTION__,
                                               g_strerror(errno)));
        }

        if (valid & ATTR_UID) { 
                rc = lchown(path, uid, buf.st_gid);
                if (rc)
                        out_save_errno(g_error("%s: setattr chown: %s",
                                               __FUNCTION__,
                                               g_strerror(errno)));
                if ( lstat(path, &buf) )
                        out_save_errno(g_error("%s: setattr chown(stat): %s",
                                               __FUNCTION__,
                                               g_strerror(errno))); 
        }

        if (valid & ATTR_GID) { 
                rc = lchown(path, buf.st_uid, gid);
                if (rc)
                        out_save_errno(g_error("%s: setattr chgrp: %s",
                                               __FUNCTION__,
                                               g_strerror(errno))); 
                if (lstat(path, &buf))
                        out_save_errno(g_error("%s: setattr chgrp(stat): %s",
                                               __FUNCTION__,
                                               g_strerror(errno))); 
        }

        if (valid & ATTR_SIZE && !S_ISLNK(buf.st_mode) ) { 
                rc = truncate(path, size);
                if (rc)
                        out_save_errno(g_error("%s: setattr truncate: %s",
                                               __FUNCTION__,
                                               g_strerror(errno))); 
        }

        /* FIXME: We don't/can't update ctime. */
        if (valid & ATTR_MTIME) { 
                tv[0].tv_usec = 0;
                tv[1].tv_usec = 0;
                tv[0].tv_sec = mtime;
                tv[1].tv_sec = mtime;
                rc = utimes(path, tv);
                if (rc)
                        out_save_errno(g_error("%s: setattr utime: %s",
                                               __FUNCTION__,
                                               g_strerror(errno)));
        }

 out:
        return rc;
}

static void addpath(char *path, int plen, char *add, int alen)
{
        /* Make sure there's exactly one slash between the path and
           the addition. */
        if (plen > 0) {
                if (path[plen - 1] == '/') {
                        while (alen > 0 && add[0] == '/') {
                                alen --;
                                add ++;
                        }
                } else {
                        if (alen > 0 && add[0] != '/')
                                path[plen ++] = '/';
                }
        } else {
                while (alen > 0 && add[0] == '/') {
                        alen --;
                        add ++;
                }
        }

        if (alen > 0) {
                memcpy (path + plen, add, alen);
                plen += alen;
        }
        path[plen] = '\0';
}

static void makepath(char *buf, struct kml_rec *rec, char *mtpt)
{
        int mlen = strlen(mtpt);
        memcpy (buf, mtpt, mlen);
        addpath (buf, mlen, rec->path, rec->pathlen);
}

static void maketarget(char *buf, struct kml_rec *rec, char *mtpt)
{
        int mlen = strlen (mtpt);
        memcpy (buf, mtpt, mlen);
        addpath (buf, mlen, rec->target, rec->targetlen);
}


static void parent(char *path)
{
        char *lastslash;
        int plen;

        plen = strlen(path);
        while (plen > 1 && path[plen - 1] == '/')
                path[-- plen] = '\0';

        /* FIXME: This does not work for paths having or needing `..' */
        lastslash = strrchr(path, '/');
        lastslash[1] = '\0';
}


static int touch_parent(struct presto_version *ver, char *path)
{
        char ppath[MAXPATHLEN];
        strcpy(ppath, path);
        parent(ppath);
        return setattr(ppath, ATTR_MTIME, 0, 0, 0, ver->pv_mtime, 0, 0);
}


static int ensure_parent(char *path)
{
        int rc;
        char ppath[MAXPATHLEN];
        strcpy(ppath, path);
        parent(ppath);

        if (! mkdir(ppath, 0777) || errno == EEXIST)
                return 0;

        rc = ensure_parent(ppath);
        if (!rc)
                rc = mkdir(ppath,0777);
        return rc;
}


static int reint_setattr(struct kml_rec *rec, char *mtpt) 
{
        int rc;
        char path[MAXPATHLEN];
        struct stat buf;

        makepath(path, rec, mtpt);

        rc = reint_pre_secure(rec, mtpt);
        if (rc != 0)
                goto out;
        
        if (reint_check_path(path))
                out_errno(EPERM);

        /* check for update/delete conflict */
        if (lstat(path, &buf)) {
                g_error("%s: update/delete conflict: %s", __FUNCTION__,
                        g_strerror(errno));
                /* return handle_conflict ... */
                out_errno(ENOENT);
        }

        /* check for update/update conflict */
        if (buf.st_mtime != rec->old_objectv->pv_mtime) {
                g_error("%s: update/update conflict: %s", __FUNCTION__,
                        g_strerror(errno)); 
                /* return handle_conflict ... */
                out_errno(ENOENT);
        }

        rc = setattr(path, rec->valid, rec->mode, rec->uid, rec->gid,
                     rec->mtime, rec->ctime, rec->size);

 out:
        reint_post_secure();
        return rc;
}


static int reint_create(struct kml_rec *rec, char *mtpt)
{
        int rc, exists = 0;
        struct stat buf;
        char path[MAXPATHLEN];

        makepath(path, rec, mtpt);

        rc = reint_pre_secure(rec, mtpt);
        if (rc != 0)
                goto out;
        
        if (reint_check_path(path))
                return -EPERM;

        if (lstat(path, &buf)) {
                if (errno != ENOENT)
                        out_save_errno(g_error("%s: cannot lstat: %s",
                                               __FUNCTION__,
                                               g_strerror(errno)));
        } else {
                /* check for update/update conflict */
                if (!S_ISREG(buf.st_mode)) { 
                        g_error("object %s exists, but is not a file\n", path);
                        /* return handle_conflict ... */
                        out_errno(EEXIST);
                }

                /* check for update/update conflict */
                if (buf.st_mtime != rec->new_objectv->pv_mtime) {
                        g_error("%s: update/update conflict: %s",
                                __FUNCTION__, g_strerror(errno));
                        /* return handle_conflict ... */
                        out_errno(EEXIST);
                }
                exists = 1;
        }

        if (!exists) {
                rc = ensure_parent(path);
                if (rc)
                        goto out;
                rc = mknod(path, rec->mode, 0);
                if (rc)
                        out_errno(errno);
        }

        rc = setattr(path,  ATTR_CTIME | ATTR_MTIME | ATTR_UID | ATTR_GID, 
                     0, rec->uid, rec->gid, rec->new_objectv->pv_mtime,
                     rec->new_objectv->pv_ctime, 0); 
        if (rc)
                goto out;

        rc = touch_parent(rec->new_parentv, path);

 out:
        reint_post_secure();
        return rc;
}


static int reint_mkdir(struct kml_rec *rec, char *mtpt)
{
        int rc, exists = 0;
        struct stat buf;
        char path[MAXPATHLEN];

        makepath(path, rec, mtpt);

        rc = reint_pre_secure(rec, mtpt);
        if (rc != 0)
                goto out;
        
        if (reint_check_path(path))
                out_errno(EPERM);

        if (lstat(path, &buf)) {
                if (errno != ENOENT)
                        out_save_errno(g_error("%s: cannot lstat: %s",
                                               __FUNCTION__,
                                               g_strerror(errno)));
        } else {
                /* check for update/update conflict */
                if (!S_ISDIR(buf.st_mode)) { 
                        g_error("object %s exists, but is not a directory\n",
                                path);
                        /* return handle_conflict ... */
                        out_errno(EEXIST);
                }
                exists = 1;
        }

        if (!exists) {
                rc = ensure_parent(path);
                if (rc)
                        goto out;
                rc = mkdir(path, rec->mode);
                if (rc)
                        out_errno(errno);
        }

        rc = setattr(path,  ATTR_CTIME | ATTR_MTIME | ATTR_UID | ATTR_GID, 
                     0, rec->uid, rec->gid, rec->new_objectv->pv_mtime,
                     rec->new_objectv->pv_ctime, 0); 
        if (rc)
                goto out;

        rc = touch_parent(rec->new_parentv, path);
        
 out:
        reint_post_secure();
        return rc;
}


static int reint_mknod(struct kml_rec *rec, char *mtpt)
{
        int rc, dev, exists = 0;
        struct stat buf;
        char path[MAXPATHLEN];

        makepath(path, rec, mtpt);

        rc = reint_pre_secure(rec, mtpt);
        if (rc != 0)
                goto out;
        
        if (reint_check_path(path))
                return -EPERM;

        dev = rec->rdev ? rec->rdev : makedev(rec->major, rec->minor);

        if (lstat(path, &buf)) {
                if (errno != ENOENT)
                        out_errno(errno);
        } else {
                if (S_ISBLK(rec->mode)) {
                        if (!S_ISBLK(buf.st_mode) || buf.st_rdev != dev) {
                                g_error("object %s exists, but is not a %d:%d block device\n", path, major(dev), minor(dev));
                                out_errno(EEXIST);
                        }
                } else if (S_ISCHR(rec->mode)) {
                        if (!S_ISCHR(buf.st_mode) || buf.st_rdev != dev) {
                                g_error("object %s exists, but is not a %d:%d character device\n", path, major(dev), minor(dev));
                                out_errno(EEXIST);
                        }
                } else if (S_ISFIFO(rec->mode)) {
                        if (!S_ISFIFO(buf.st_mode)) {
                                g_error("object %s exists, but is not a FIFO\n", path);
                                out_errno(EEXIST);
                        }
                } else {
                        g_error("object %s is not a known node type\n",
                                path);
                        out_errno(EEXIST);
                }
                exists = 1;
        }

        if (!exists) {
                rc = ensure_parent(path);
                if (rc)
                        goto out;
                rc = mknod(path, rec->mode, dev);
                if (rc)
                        out_errno(errno);
        }

        rc = setattr(path,  ATTR_CTIME | ATTR_MTIME | ATTR_UID | ATTR_GID, 
                     0, rec->uid, rec->gid, rec->new_objectv->pv_mtime,
                     rec->new_objectv->pv_ctime, 0); 
        if (rc)
                goto out;

        rc = touch_parent(rec->new_parentv, path);

 out:
        reint_post_secure();
        return rc;
}


static int reint_noop(struct kml_rec *rec, char *mtpt)
{
        return 0;
}


static int reint_symlink(struct kml_rec *rec, char *mtpt)
{
        int rc, exists = 0;
        struct stat buf;
        char path[MAXPATHLEN];
        char target[MAXPATHLEN];

        rc = reint_pre_secure(rec, mtpt);
        if (rc != 0)
                goto out;
        
        makepath(path, rec, mtpt);
        if (reint_check_path(path))
                out_errno(EPERM);

        memcpy (target, rec->target, rec->targetlen);
        target[rec->targetlen] = '\0';

        /* check for update/update conflict */
        rc = lstat(path, &buf);
        if (rc) {
                if (errno != ENOENT)
                        out_errno(errno);
        } else if (S_ISLNK(buf.st_mode)) {
                char oldtarget[MAXPATHLEN];

                memset(oldtarget, 0, MAXPATHLEN);
                readlink(path, oldtarget, MAXPATHLEN - 1);

                if (strcmp(target, oldtarget)) {
                        g_error("object %s is not a symlink to %s\n",
                                path, target);
                        /* return handle_conflict ... */
                        out_errno(EEXIST);
                }
                exists = 1;
        } else {
                g_error("object %s exists, but is not a symlink\n", path);
                /* return handle_conflict ... */
                out_errno(EEXIST);
        }

        if (!exists) {
                rc = ensure_parent(path);
                if (rc)
                        goto out;
                rc = symlink(target, path);
                if (rc)
                        out_errno(errno);
        }

        rc = touch_parent(rec->new_parentv, path);

 out:
        reint_post_secure();
        return rc;
}


static int reint_close(struct kml_rec *rec, char *mtpt)
{
        int rc;
        char path[MAXPATHLEN];
        
        rc = reint_pre_secure(rec, mtpt);
        if (rc != 0)
                goto out;
        
        makepath(path, rec, mtpt);
        if (reint_check_path(path))
                out_errno(EPERM);

        /* FIXME: Add callback to transfer file data. */
        rc = setattr(path,
                     ATTR_CTIME | ATTR_MTIME | ATTR_UID | ATTR_GID | ATTR_SIZE, 
                     0, rec->uid, rec->gid, rec->new_objectv->pv_mtime,
                     rec->new_objectv->pv_ctime, rec->new_objectv->pv_size);

 out:
        reint_post_secure();
        return rc;
}


static int reint_unlink(struct kml_rec *rec, char *mtpt)
{
        int rc, exists = 0;
        struct stat buf;
        char path[MAXPATHLEN];

        rc = reint_pre_secure(rec, mtpt);
        if (rc != 0)
                goto out;
                
        makepath(path, rec, mtpt);
        strncat (path, rec->target, rec->targetlen);
        if (reint_check_path(path))
                out_errno(EPERM);

        /* check for delete/update conflict */
        if (lstat(path, &buf)) {
                if (errno != ENOENT) {
                        g_error("%s: cannot lstat: %s",
                                __FUNCTION__, g_strerror(errno));
                        /* return handle_conflict ... */
                        out_errno(EEXIST);
                }
        } else
                exists = 1;

        if (exists) {
                rc = unlink(path);
                if (rc)
                        out_errno(errno);
        }

        rc = touch_parent(rec->new_parentv, path);

 out:
        reint_post_secure();
        return rc;
}


static int reint_link(struct kml_rec *rec, char *mtpt)
{
        int rc, exists = 0;
        struct stat obuf, buf;
        char oldname[MAXPATHLEN];
        char newname[MAXPATHLEN];

        rc = reint_pre_secure(rec, mtpt);
        if (rc != 0)
                goto out;

        makepath(oldname, rec, mtpt);
        maketarget(newname, rec, mtpt);
        if (reint_check_path(newname))
                out_errno(EPERM);

        /* check for update/update conflict */
        rc = lstat(newname, &buf);
        if (rc) {
                if (errno != ENOENT)
                        out_errno(errno);
        } else if ((rc = lstat(oldname, &obuf))) {
                out_errno(errno);
        } else {
                if (obuf.st_dev != buf.st_dev || obuf.st_ino != buf.st_ino) {
                        g_error("object %s is not a link to %s\n",
                                newname, oldname);
                        /* return handle_conflict ... */
                        out_errno(EEXIST);
                }
                exists = 1;
        }

        if (!exists) {
                rc = ensure_parent(newname);
                if (rc)
                        goto out;
                rc = link(oldname, newname);
                if (rc)
                        out_errno(errno);
        }

        rc = touch_parent(rec->new_parentv, newname);

 out:
        reint_post_secure();
        return rc;
}


static int reint_rename(struct kml_rec *rec, char *mtpt)
{
        int rc;
        struct stat buf;
        char oldname[MAXPATHLEN];
        char newname[MAXPATHLEN];

        rc = reint_pre_secure(rec, mtpt);
        if (rc != 0)
                goto out;

        makepath(oldname, rec, mtpt);
        if (reint_check_path(oldname))
                out_errno(EPERM);

        maketarget(newname, rec, mtpt);
        if (reint_check_path(newname))
                out_errno(EPERM);

        /* check for update/update conflict */
        rc = lstat(newname, &buf);
        if (rc) {
                if (errno != ENOENT)
                        out_errno(errno);
        } else {
                g_error("object %s already exists\n", newname);
                out_errno(EEXIST);
        }

        rc = ensure_parent(newname);
        if (rc)
                goto out;
        rc = rename(oldname, newname);
        if (rc)
                out_errno(errno);

        rc = touch_parent(rec->new_objectv, oldname);
        if (!rc)
                rc = touch_parent(rec->new_objectv, newname);

 out:
        reint_post_secure();
        return rc;
}


static int reint_rmdir(struct kml_rec *rec, char *mtpt)
{
        int rc, exists = 0;
        struct stat buf;
        char path[MAXPATHLEN];

        rc = reint_pre_secure(rec, mtpt);
        if (rc != 0)
                goto out;
                
        makepath(path, rec, mtpt);
        strncat(path, rec->target, rec->targetlen);
        if (reint_check_path(path))
                out_errno(EPERM);

        /* check for delete/update conflict */
        if (lstat(path, &buf)) {
                if (errno != ENOENT) {
                        g_error("%s: cannot lstat: %s",
                                __FUNCTION__, g_strerror(errno));
                        /* return handle_conflict ... */
                        out_errno(EEXIST);
                }
        } else
                exists = 1;

        if (exists) {
                rc = rmdir(path);
                if (rc)
                        out_errno(errno);
        }

        rc = touch_parent(rec->new_parentv, path);

 out:
        reint_post_secure();
        return rc;
}


typedef int (*fsreinter)(struct kml_rec *rec, char *basedir);

static fsreinter fsreinters[KML_OPCODE_NUM] =
{
        [KML_OPCODE_CLOSE] = reint_close,
        [KML_OPCODE_CREATE] = reint_create,
        [KML_OPCODE_LINK] = reint_link,
        [KML_OPCODE_MKDIR] = reint_mkdir,
        [KML_OPCODE_MKNOD] = reint_mknod,
        [KML_OPCODE_NOOP] = reint_noop,
        [KML_OPCODE_RENAME] = reint_rename,
        [KML_OPCODE_RMDIR] = reint_rmdir,
        [KML_OPCODE_SETATTR] = reint_setattr,
        [KML_OPCODE_SYMLINK] = reint_symlink,
        [KML_OPCODE_UNLINK] = reint_unlink,
};

int kml_fsreint(struct kml_rec *rec, char *basedir)
{
        int op = rec->prefix.hdr->opcode;
        fsreinter reinter = (op < 0 ||
                             op >= sizeof (fsreinters) / sizeof (*fsreinters))
                ? NULL : fsreinters[op];

        if (!reinter)
                return -EINVAL;

        return reinter(rec, basedir);
}
