/* -*- mode: c; c-basic-offset: 8; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=8:tabstop=8:
 *
 *  Copyright (C) 2002 Cluster File Systems, Inc.
 *                     Author: Gordon Matzigkeit <gord@fig.org>
 *
 *   This file is part of InterMezzo, http://www.inter-mezzo.org.
 *
 *   InterMezzo is free software; you can redistribute it and/or
 *   modify it under the terms of version 2 of the GNU General Public
 *   License as published by the Free Software Foundation.
 *
 *   InterMezzo is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with InterMezzo; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Inversion of KML records.
 *
 */

#define _GNU_SOURCE 1
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>


#include <linux/intermezzo_lib.h>
#include <glib.h>
#include <izolib.h>

/* Release memory allocated by kml_invert. */
void kml_invert_free(struct kml_rec *rec)
{
        if (rec == NULL)
                return;

#define do_free(var)                            \
        if (var) {                              \
                free(var);                      \
                var = NULL;                     \
        }

        do_free(rec->prefix.hdr);
        do_free(rec->path);
        do_free(rec->name);
        do_free(rec->target);
        do_free(rec->old_objectv);
        do_free(rec->new_objectv);
        do_free(rec->old_parentv);
        do_free(rec->new_parentv);
        do_free(rec->old_targetv);
        do_free(rec->new_targetv);

#undef do_free
}


/* Join a path and target into a path. */
static void join_path_target(struct kml_rec *src, struct kml_rec *dst)
{
        int size = src->pathlen + src->targetlen;
        int slash = 0;

        if (src->pathlen == 0 ||
            (src->path[src->pathlen - 1] != '/' && src->targetlen > 0)) {
                slash = 1;
                size ++;
        }

        dst->path = g_new(char, size);
        if (dst->path == NULL)
                return;

        if (src->pathlen != 0)
                memcpy (dst->path, src->path, src->pathlen);
        dst->pathlen = src->pathlen;

        if (slash) {
                dst->path[dst->pathlen] = '/';
                dst->pathlen++;
        }

        memcpy (dst->path + dst->pathlen, src->target, src->targetlen);
        dst->pathlen += src->targetlen;
}


#if !defined(__GLIBC__) || (__GLIBC__ < 2) || (__GLIBC__ == 2 && __GLIBC_MINOR__ < 2)
/* A replacement for missing memrchr. */
static void *memrchr(void *s, int c, size_t n)
{
        char *p = s;
        while (n > 0)
                if (p[-- n] == c)
                        return p + n;
        return NULL;
}
#endif


static void *copy_mem(void *ptr, size_t size)
{
        void *ret = NULL;
        if (size > 0) {
                ret = g_new(char, size);
                if (ret)
                        memcpy(ret, ptr, size);
        }
        return ret;
}


/* Break up the path into parent directory and target. */
static void split_path(struct kml_rec *src, struct kml_rec *dst)
{
        char *p = memrchr(src->path, '/', src->pathlen);

        if (p) {
                if (p - src->path > 0)
                        dst->pathlen = p - src->path;
                else
                        dst->pathlen = 1;
                dst->path = copy_mem(src->path, dst->pathlen);
                p ++;
        } else {
                p = src->path;
                dst->pathlen = 1;
                dst->path = g_new(char, 1);
                dst->path[0] = '/';
        }

        dst->targetlen = src->pathlen - (p - src->path);
        dst->target = copy_mem(p, dst->targetlen);
}

static struct presto_version *copy_ver(struct presto_version *ver)
{
        struct presto_version *nver;
        if (ver == NULL)
                return NULL;

        nver = copy_mem(ver, sizeof(*ver));
        return nver;
}


struct close_node {
        GHashTable *ents;
        /* VER is nonzero when a close is needed on this node. */
        struct presto_version *ver;
};

static GHashTable *close_map = NULL;

static int kml_invert_close(struct kml_rec *rec, struct kml_rec *inv_rec)
{
        int i = 0;
        char *name = NULL;
        GHashTable *cdir = close_map;

        /* Set the attributes. */
        if (rec->old_objectv) {
                inv_rec->prefix.hdr->opcode = KML_OPCODE_SETATTR;
                inv_rec->pathlen = rec->pathlen;
                inv_rec->path = copy_mem(rec->path, inv_rec->pathlen);
                inv_rec->old_objectv = copy_ver(rec->new_objectv);
                inv_rec->valid = ATTR_MTIME | ATTR_CTIME | ATTR_SIZE;
                inv_rec->ctime = rec->old_objectv->pv_mtime;
                inv_rec->mtime = rec->old_objectv->pv_ctime;
                inv_rec->size = rec->old_objectv->pv_size;
        } else
                inv_rec->prefix.hdr->opcode = KML_OPCODE_NOOP;
        
        /* Record that we need a fetch for this file. */
        while (cdir && i < rec->pathlen) {
                if (name) {
                        int end = -1;
                        if (rec->path[i] == '/')
                                end = i - 1;
                        else if (i == rec->pathlen - 1)
                                end = i;

                        if (end >= 0) {
                                struct close_node *cn;
                                int namelen = rec->path + end - name + 1;
                                char *name0 = g_new(char, namelen + 1);

                                memcpy(name0, name, namelen);
                                name0[namelen] = '\0';

                                cn = g_hash_table_lookup(cdir, name0);
                                if (cn == NULL) {
                                        cn = g_new0(struct close_node, 1);
                                        g_hash_table_insert(cdir, name0, cn);
                                } else {
                                        free(name0);
                                }

                                if (i == rec->pathlen - 1)
                                        cn->ver = copy_ver(rec->old_objectv);
                                else if (cn->ents == NULL)
                                        cn->ents =
                                                g_hash_table_new(g_str_hash,
                                                                 g_str_equal);
                                cdir = cn->ents;
                                name = NULL;
                        }
                } else if (rec->path[i] != '/')
                        name = rec->path + i;
                i++;
        }
        return 0;
}

static int kml_invert_create(struct kml_rec *rec, struct kml_rec *inv_rec)
{
        inv_rec->prefix.hdr->opcode = KML_OPCODE_UNLINK;
        split_path (rec, inv_rec);
        inv_rec->old_objectv = copy_ver(rec->new_objectv);
        inv_rec->new_parentv = copy_ver(rec->old_parentv);
        inv_rec->old_parentv = copy_ver(rec->new_parentv);
        return 0;
}

static int kml_invert_link(struct kml_rec *rec, struct kml_rec *inv_rec)
{
        struct kml_rec tmp_rec;
        inv_rec->prefix.hdr->opcode = KML_OPCODE_UNLINK;
        inv_rec->old_targetlen = rec->pathlen;
        inv_rec->old_target = copy_mem(rec->path, rec->pathlen);

        /* Juggle things so that the target gets stored in the inverse
           path. */
        tmp_rec.pathlen = rec->targetlen;
        tmp_rec.path = rec->target;
        split_path (&tmp_rec, inv_rec);

        inv_rec->old_objectv = copy_ver(rec->new_objectv);
        inv_rec->new_parentv = copy_ver(rec->old_parentv);
        inv_rec->old_parentv = copy_ver(rec->new_parentv);
        return 0;
}

static int kml_invert_mkdir(struct kml_rec *rec, struct kml_rec *inv_rec)
{
        inv_rec->prefix.hdr->opcode = KML_OPCODE_RMDIR;
        split_path (rec, inv_rec);
        inv_rec->old_objectv = copy_ver(rec->new_objectv);
        inv_rec->new_parentv = copy_ver(rec->old_parentv);
        inv_rec->old_parentv = copy_ver(rec->new_parentv);
        inv_rec->old_mode = rec->mode;
        inv_rec->old_uid = rec->uid;
        inv_rec->old_gid = rec->gid;
        return 0;
}

static int kml_invert_mknod(struct kml_rec *rec, struct kml_rec *inv_rec)
{
        inv_rec->prefix.hdr->opcode = KML_OPCODE_UNLINK;
        split_path (rec, inv_rec);
        inv_rec->old_objectv = copy_ver(rec->new_objectv);
        inv_rec->new_parentv = copy_ver(rec->old_parentv);
        inv_rec->old_parentv = copy_ver(rec->new_parentv);
        inv_rec->old_mode = rec->mode;
        inv_rec->old_rdev = rec->rdev ? rec->rdev :
                makedev(rec->major, rec->minor);
        inv_rec->old_uid = rec->uid;
        inv_rec->old_gid = rec->gid;
        return 0;
}

static int kml_invert_noop(struct kml_rec *rec, struct kml_rec *inv_rec)
{
        inv_rec->prefix.hdr->opcode = KML_OPCODE_NOOP;
        return 0;
}

static int kml_invert_rename(struct kml_rec *rec, struct kml_rec *inv_rec)
{
        int i = 0;
        char *name = NULL;
        GHashTable *cdir = close_map;

        inv_rec->prefix.hdr->opcode = KML_OPCODE_RENAME;

        inv_rec->pathlen = rec->targetlen;
        inv_rec->path = copy_mem(rec->target, rec->targetlen);
        inv_rec->targetlen = rec->pathlen;
        inv_rec->target = copy_mem(rec->path, rec->pathlen);

        /* Adjust our close map so we fetch correct filenames. */
        while (cdir && i < inv_rec->pathlen) {
                if (name) {
                        int end = -1;
                        if (inv_rec->path[i] == '/')
                                end = i - 1;
                        else if (i == inv_rec->pathlen - 1)
                                end = i;

                        if (end >= 0) {
                                struct close_node *cn;
                                int namelen = inv_rec->path + end - name + 1;
                                char *name0 = g_new(char, namelen + 1);

                                memcpy(name0, name, namelen);
                                name0[namelen] = '\0';

                                cn = g_hash_table_lookup(cdir, name0);
                                if (cn == NULL) {
                                        free (name0);
                                        break;
                                }

                                if (i == inv_rec->pathlen - 1) {
                                        g_hash_table_insert(cdir, name0, cn);
                                        g_hash_table_remove(cdir, name0);
                                } else {
                                        cdir = cn->ents;
                                        free (name0);
                                }
                                name = NULL;
                        }
                } else if (inv_rec->path[i] != '/')
                        name = inv_rec->path + i;
                i++;
        }

        inv_rec->new_objectv = copy_ver(rec->old_objectv);
        inv_rec->old_objectv = copy_ver(rec->new_objectv);
        inv_rec->new_parentv = copy_ver(rec->old_parentv);
        inv_rec->old_parentv = copy_ver(rec->new_parentv);
        return 0;
}


static int kml_invert_rmdir(struct kml_rec *rec, struct kml_rec *inv_rec)
{
        inv_rec->prefix.hdr->opcode = KML_OPCODE_MKDIR;
        join_path_target(rec, inv_rec);
        inv_rec->new_objectv = copy_ver(rec->old_objectv);
        inv_rec->new_parentv = copy_ver(rec->old_parentv);
        inv_rec->old_parentv = copy_ver(rec->new_parentv);
        inv_rec->mode = rec->old_mode;
        inv_rec->uid = rec->old_uid;
        inv_rec->gid = rec->old_gid;
        return 0;
}

static int kml_invert_setattr(struct kml_rec *rec, struct kml_rec *inv_rec)
{
        inv_rec->prefix.hdr->opcode = KML_OPCODE_SETATTR;
        join_path_target(rec, inv_rec);
        inv_rec->old_objectv = g_new(struct presto_version, 1);
        inv_rec->old_objectv->pv_ctime = rec->ctime;
        inv_rec->old_objectv->pv_mtime = rec->mtime;
        inv_rec->old_objectv->pv_size = rec->size;
        inv_rec->valid = rec->valid;
        if (rec->old_objectv) {
                inv_rec->ctime = rec->old_objectv->pv_ctime;
                inv_rec->mtime = rec->old_objectv->pv_mtime;
        }
        inv_rec->old_mode = rec->mode;
        inv_rec->mode = rec->old_mode;
        inv_rec->old_uid = rec->uid;
        inv_rec->uid = rec->old_uid;
        inv_rec->old_gid = rec->gid;
        inv_rec->gid = rec->old_gid;
        return 0;
}

static int kml_invert_symlink(struct kml_rec *rec, struct kml_rec *inv_rec)
{
        inv_rec->prefix.hdr->opcode = KML_OPCODE_UNLINK;
        split_path (rec, inv_rec);
        inv_rec->old_targetlen = rec->targetlen;
        inv_rec->old_target = copy_mem(rec->target, rec->targetlen);

        inv_rec->old_objectv = copy_ver(rec->new_objectv);
        inv_rec->new_parentv = copy_ver(rec->old_parentv);
        inv_rec->old_parentv = copy_ver(rec->new_parentv);
        inv_rec->old_mode = S_IFLNK;
        return 0;
}

static int kml_invert_unlink_create(struct kml_rec *rec,
                                    struct kml_rec *inv_rec)
{
        inv_rec->prefix.hdr->opcode = KML_OPCODE_CREATE;
        join_path_target(rec, inv_rec);

        inv_rec->new_objectv = copy_ver(rec->old_objectv);
        inv_rec->new_parentv = copy_ver(rec->old_parentv);
        inv_rec->old_parentv = copy_ver(rec->new_parentv);
        inv_rec->mode = rec->old_mode;
        inv_rec->uid = rec->old_uid;
        inv_rec->gid = rec->old_gid;
        return 0;
}

static int kml_invert_unlink_link(struct kml_rec *rec,
                                  struct kml_rec *inv_rec)
{
        inv_rec->prefix.hdr->opcode = KML_OPCODE_LINK;
        join_path_target(rec, inv_rec);
        inv_rec->targetlen = rec->old_targetlen;
        inv_rec->target = copy_mem(rec->old_target, rec->targetlen);

        inv_rec->new_objectv = copy_ver(rec->old_objectv);
        inv_rec->new_parentv = copy_ver(rec->old_parentv);
        inv_rec->old_parentv = copy_ver(rec->new_parentv);
        return 0;
}

static int kml_invert_unlink_mknod(struct kml_rec *rec,
                                   struct kml_rec *inv_rec)
{
        inv_rec->prefix.hdr->opcode = KML_OPCODE_MKNOD;
        join_path_target(rec, inv_rec);

        inv_rec->new_objectv = copy_ver(rec->old_objectv);
        inv_rec->new_parentv = copy_ver(rec->old_parentv);
        inv_rec->old_parentv = copy_ver(rec->new_parentv);
        inv_rec->mode = rec->old_mode;
        inv_rec->uid = rec->old_uid;
        inv_rec->gid = rec->old_gid;
#if 0
        /* FIXME: This field isn't being packed/unpacked. */
        inv_rec->rdev = rec->old_rdev;
#endif
        inv_rec->major = major(rec->rdev);
        inv_rec->minor = minor(rec->rdev);
        return 0;
}

static int kml_invert_unlink_symlink(struct kml_rec *rec,
                                     struct kml_rec *inv_rec)
{
        inv_rec->prefix.hdr->opcode = KML_OPCODE_SYMLINK;
        join_path_target(rec, inv_rec);
        inv_rec->targetlen = rec->old_targetlen;
        inv_rec->target = copy_mem(rec->old_target, rec->targetlen);

        inv_rec->new_objectv = copy_ver(rec->old_objectv);
        inv_rec->new_parentv = copy_ver(rec->old_parentv);
        inv_rec->old_parentv = copy_ver(rec->new_parentv);
        inv_rec->uid = rec->old_uid;
        inv_rec->gid = rec->old_gid;
        return 0;
}

static int kml_invert_unlink(struct kml_rec *rec, struct kml_rec *inv_rec)
{
        if (S_ISLNK(rec->old_mode))
                return kml_invert_unlink_symlink(rec, inv_rec);

        if (rec->old_targetlen != 0)
                return kml_invert_unlink_link(rec, inv_rec);

        if (S_ISREG(rec->old_mode))
                return kml_invert_unlink_create(rec, inv_rec);

        return kml_invert_unlink_mknod(rec, inv_rec);
}

typedef int (*inverter)(struct kml_rec *rec, struct kml_rec *inv_rec);

static inverter inverters[KML_OPCODE_NUM] =
{
        [KML_OPCODE_CLOSE] = kml_invert_close,
        [KML_OPCODE_CREATE] = kml_invert_create,
        [KML_OPCODE_LINK] = kml_invert_link,
        [KML_OPCODE_MKDIR] = kml_invert_mkdir,
        [KML_OPCODE_MKNOD] = kml_invert_mknod,
        [KML_OPCODE_NOOP] = kml_invert_noop,
        [KML_OPCODE_RENAME] = kml_invert_rename,
        [KML_OPCODE_RMDIR] = kml_invert_rmdir,
        [KML_OPCODE_SETATTR] = kml_invert_setattr,
        [KML_OPCODE_SYMLINK] = kml_invert_symlink,
        [KML_OPCODE_UNLINK] = kml_invert_unlink,
};

struct kml_rec *kml_invert(struct kml_rec *rec, int recno)
{
        int op = rec->prefix.hdr->opcode;
        inverter inv;
        struct kml_rec *inv_rec;

        if (close_map == NULL)
                close_map = g_hash_table_new(g_str_hash, g_str_equal);

        inv_rec = g_new0(struct kml_rec, 1);
        inv_rec->prefix.hdr = copy_mem(rec->prefix.hdr,
                                       sizeof(*rec->prefix.hdr));

        inv_rec->prefix.groups =
                copy_mem (rec->prefix.groups,
                          inv_rec->prefix.hdr->ngroups *
                          sizeof(*rec->prefix.groups));

        inv_rec->valid = rec->valid;

        inv_rec->suffix = g_new(struct kml_suffix, 1);
        inv_rec->suffix->prevrec = recno - 1;
        inv_rec->suffix->recno = recno;
        inv_rec->suffix->time = rec->suffix->time;

        inv = (op < 0 || op >= sizeof(inverters) / sizeof(*inverters))
                ? NULL : inverters[op];

        if (inv == NULL) {
                kml_invert_free(inv_rec);
                return NULL;
        }

        inv(rec, inv_rec);
        return inv_rec;
}

struct close_status {
        int status;
        char *path;
};

static void do_close(gpointer key, gpointer value, gpointer user_data)
{
        char *name = key;
        struct close_node *cn = value;
        struct close_status *cs = user_data;
        char fullname[strlen(cs->path) + 1 + strlen(name) + 1];
        struct close_status ncs = { cs->status, fullname };

        strcpy(fullname, cs->path);
        strcat(fullname, "/");
        strcat(fullname, name);

        if (cn->ver)
                g_message("FIXME: would do close on: %s\n", fullname);
        if (cn->ents)
                g_hash_table_foreach (cn->ents, do_close, &ncs);
}

/* Perform the deferred closes. */
int kml_invert_closes(void)
{
        struct close_status cs = { 0, "" };
        if (close_map != NULL)
                g_hash_table_foreach(close_map, do_close, &cs);
        return cs.status;
}
